#include <gtest_extension.hpp>
#include <iostream>
#include <sstream>
#include <limits>
#include <type_traits>
#include <cmath>

#include "komplex.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

struct KomplexTeszt {
    double re, im;
};

std::ostream& operator<<(std::ostream& os, KomplexTeszt const& k) {
    os << k.re;
    if (k.im >= 0)
    os << "+";
    os << k.im << "i";
    return os;
}

bool operator==(KomplexTeszt const& c1, Komplex const& c2) {
    static_assert(sizeof(Komplex) == sizeof(KomplexTeszt) && std::is_standard_layout<Komplex>::value,
            "Ne változtasd meg a Komplex adattagjait!");
    KomplexTeszt const& c22 = reinterpret_cast<KomplexTeszt const&>(c2);
    double e = std::numeric_limits<float>::epsilon(); // a float-tal megengedők vagyunk
    return std::fabs(c1.re - c22.re) < e && std::fabs(c1.im - c22.im) < e;
}

template<class T>
using van_default_konstruktor = decltype(T());

MY_TEST(KomplexTeszt, f01_1_default_konstruktor, Komplex, van_default_konstruktor)
{
    Komplex c1;
    EXPECT_EQ((KomplexTeszt{0.0, 0.0}), c1);
}

template<class T>
using van_egyparameteres_konstruktor = decltype(T(0.0));

MY_TEST(KomplexTeszt, f01_2_egyparameteres_konstruktor, Komplex, van_egyparameteres_konstruktor)
{
    Komplex c1(4.5);
    EXPECT_EQ((KomplexTeszt{4.5, 0.0}), c1);

    Komplex c2 = 3.5; // Legyen implicit konverzió!
    EXPECT_EQ((KomplexTeszt{3.5, 0.0}), c2);
}

template<class T>
using van_ketparameteres_konstruktor = decltype(T(0.0, 0.0));

MY_TEST(KomplexTeszt, f01_3_ketparameteres_konstruktor, Komplex, van_ketparameteres_konstruktor)
{
    Komplex c1(4.5, 3.7);
    EXPECT_EQ((KomplexTeszt{4.5, 3.7}), c1);
}

template<class T>
using vannak_getterek = decltype(T().get_re(), T().get_im());

MY_TEST(KomplexTeszt, f02_getterek, Komplex, vannak_getterek)
{
    const Komplex c1(4.5, 3.7);
    double re = c1.get_re(); // .get_re() és .get_im() legyen hívható konstans Komplexen is!
    double im = c1.get_im();
    EXPECT_DOUBLE_EQ(4.5, re);
    EXPECT_DOUBLE_EQ(3.7, im);
}

template<class T>
using van_operator_cout = decltype(std::cout << T());

MY_TEST(KomplexTeszt, f03_operator_cout, Komplex, van_operator_cout)
{
    auto Komplex_to_string = [](Komplex const& c) {
        std::ostringstream oss;
        oss << c;
        return oss.str();
    };
    EXPECT_EQ("3+4i", Komplex_to_string(Komplex(3, 4)));
    EXPECT_EQ("3-4i", Komplex_to_string(Komplex(3, -4)));
    EXPECT_EQ("-3+4i", Komplex_to_string(Komplex(-3, 4)));
    EXPECT_EQ("-3-4i", Komplex_to_string(Komplex(-3, -4)));
}

template<class T>
using van_operator_plusz = decltype(T() + T());

MY_TEST(KomplexTeszt, f04_1_operator_plusz, Komplex, van_operator_plusz)
{
    EXPECT_EQ((KomplexTeszt{4, 6}), Komplex(3, 4) + Komplex(1, 2));              // Komplex + Komplex
    EXPECT_EQ((KomplexTeszt{7, 2}), 6.0 + Komplex(1, 2));                        // double + Komplex
    EXPECT_EQ((KomplexTeszt{4, 6}), operator+(Komplex(3, 4), Komplex(1, 2)));    // globális az operator+
}

template<class T>
using van_operator_csillag = decltype(T() * T());

MY_TEST(KomplexTeszt, f04_2_operator_csillag, Komplex, van_operator_csillag)
{
    EXPECT_EQ((KomplexTeszt{-5, 10}), Komplex(3, 4) * Komplex(1, 2));            // Komplex * Komplex
    EXPECT_EQ((KomplexTeszt{ 6, 12}), 6.0 * Komplex(1, 2));                      // double * Komplex
    EXPECT_EQ((KomplexTeszt{-5, 10}), operator*(Komplex(3, 4), Komplex(1, 2)));  // globális az operator*
}

template<class T>
using van_operator_plusz_egyenlo = decltype(std::declval<T&>() += T());

MY_TEST(KomplexTeszt, f04_3_operator_plusz_egyenlo, Komplex, van_operator_plusz_egyenlo)
{
    Komplex c(3, 4);
    EXPECT_EQ((KomplexTeszt{4, 6}),   c += Komplex(1, 2));                       // Komplex += Komplex
    EXPECT_EQ((KomplexTeszt{10, 6}),  c += 6.0);                                 // Komplex += double
    EXPECT_EQ((KomplexTeszt{11, 8}),  operator+=(c, Komplex(1, 2)));             // globális az operator+=
}

template<class T>
using van_operator_csillag_egyenlo = decltype(std::declval<T&>() *= T());

MY_TEST(KomplexTeszt, f04_4_operator_csillag_egyenlo, Komplex, van_operator_csillag_egyenlo)
{
    Komplex c(3, 4);
    EXPECT_EQ((KomplexTeszt{-5, 10}),   c *= Komplex(1, 2));                     // Komplex *= Komplex
    EXPECT_EQ((KomplexTeszt{-10, 20}),  c *= 2.0);                               // Komplex *= double
    EXPECT_EQ((KomplexTeszt{-60, 20}),  operator*=(c, Komplex(2, 2)));           // globális az operator*=
}

template<class T>
using van_operator_per = decltype(T() / 0.0);

MY_TEST(KomplexTeszt, f05_operator_per, Komplex, van_operator_per)
{
    EXPECT_EQ((KomplexTeszt{3, 2}), Komplex(6, 4) / 2.0);                        // Komplex / double
    EXPECT_EQ((KomplexTeszt{3, 2}), operator/(Komplex(6, 4), 2.0));              // globális az operator/
}

TEST(KomplexTeszt, f06_Komplex_sqrt)
{
    EXPECT_EQ((KomplexTeszt{2, 0}), Komplex_sqrt(4.0));
    EXPECT_EQ((KomplexTeszt{0, 2}), Komplex_sqrt(-4.0));
}

template<class T>
using vannak_trigonometriai_getterek = decltype(T().get_r(), T().get_fi());

MY_TEST(KomplexTeszt, f07_trigonometriai_getterek, Komplex, vannak_trigonometriai_getterek)
{
    const Komplex c1(1, 1);
    EXPECT_DOUBLE_EQ(sqrt(2), c1.get_r());
    EXPECT_DOUBLE_EQ(M_PI / 4, c1.get_fi());

    EXPECT_DOUBLE_EQ( 3*M_PI / 4, Komplex(-1,  1).get_fi());
    EXPECT_DOUBLE_EQ(  -M_PI / 4, Komplex( 1, -1).get_fi());
    EXPECT_DOUBLE_EQ(-3*M_PI / 4, Komplex(-1, -1).get_fi());
}

TEST(KomplexTeszt, f08_Komplex_from_r_fi)
{
    EXPECT_EQ((KomplexTeszt{1, 1}), Komplex_from_r_fi(sqrt(2), M_PI / 4));
}

